package BadTraffic;

use strict;
use warnings;
use DBI;
use JSON qw();
use LWP;
use Socket;
use Sys::Hostname;
use Net::CIDR::Lite;

# WARNING TO FUTURE AUTHORS:
# Puppet currently runs the "blocker" script in VERBOSE=1 mode and
# interprets the debug output it receives as IP addresses to be blocked.
# Be very careful about adding additional debug messages, esp to
# whitelist routines. We learned this the hard way when a few dozen core
# servers all blocked NCSU IPs on a puppet update on 3/25/2015.
# See puppet code installed: /var/lib/puppet/lib/facter/badtraffic.rb

require Exporter;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration  use BadTraffic ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = (
    'all' => [
        qw(
            get_cat_list
            get_fw_blocks
            get_whitelist
            get_db_blocks
            fw_block
            fw_unblock
            fw_insanity
            db_block
            db_unblock
            db_block_expire
            cat_add
            cat_remove
            wl_add
            wl_remove
            )
    ]
);

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

# no default exports needed
our @EXPORT = qw();

use version;
our $VERSION = qv('1.6.9');

=head1 NAME

BadTraffic - Perl extension for automatic firewall management.

=head1 VERSION

This is BadTraffic version 1.6.9.

=head1 SYNOPSIS

  my $bt = BadTraffic->new(
    DB_PASS => "<password>",
    DB_HOST => "<hostname>",
    DB_USER => "<username>",
    DB_NAME => "<dbname>",
    FW_TYPE => 'iptables',
  );

  my $cat = 'web-dyn';
  my $wl_cidr = $bt->get_whitelist($cat);
  my $db_cidr = $bt->get_db_blocks($cat);
  my $fw_cidr = $bt->get_fw_blocks();

  foreach my $ip (@badiplist) {
      next if ($wl_cidr->find($ip); # skip whitelist matches

      $bt->fw_block($ip);
      $bt->db_block(
        address => $ip,
        reason  => 'they are very bad',
        days    => 31,
        category => $cat
      );
  }

=head1 DESCRIPTION

BadTraffic reads and writes to a MySQL database and blocks traffic
according entries found.  It provides facilities for whitelisting
addresses or networks using CIDR formatting.


=head1 SEE ALSO

DBD::mysql and Net::CIDR::Lite are required used to greater or
lesser extent in this module.  There is also a heavy reliance on shelling
out to call /sbin/iptables and /sbin/ipfw to actually manipulate the
firewall blocks.

=head1 AUTHORS

Breandan Dezendorf, E<lt>bwdezend@ncsu.eduE<gt>

Charles J Brabec, E<lt>brabec@ncsu.eduE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2011-2013 by Breandan Dezendorf, Charles J Brabec

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.10.0 or,
at your option, any later version of Perl 5 you may have available.

=cut

=head1 PUBLIC METHODS

Each public function/method is described here.
These are how you should interact with this module.

=cut

=head2 new

 Usage    : my $bt = new BadTraffic;
 Purpose  : Create a new BadTraffic object.
 Returns  : An object in the form of a blessed hash ref.
 Argument : hash or hashref of values to set:
               DB_NAME => MySQL database name (default: badtraffic_db)
               DB_USER => MySQL username (default: autoblock)
               DB_PASS => MySQL password (default: none)
               DB_HOST => MySQL hostname (default: localhost)
               FW_TYPE => type of firewall (default: iptables)
               REST_HOST => https://sysnews.ncsu.edu
               NOCACHE => when true, will ignore cache files and make db calls
 Comments : The valid firewall types are "ipset", "iptables", 
          : "ipfw", and "route"
 
=cut

sub new {
    my $class  = shift;
    my %params = @_;
    my $self   = {
        _DB_TYPE         => 'mysql',
        _DB_USER         => 'autoblock',
        _DB_PASS         => 'mysql',
        _DB_HOST         => 'localhost',
        _DB_NAME         => 'bad_traffic',
        _FW_TYPE         => 'iptables',
        _VERBOSE         => 0,
        _DECOMPOSED_LIST => {},
        _CAT             => undef,
        _WLCIDR          => undef,
        _DBCIDR          => undef,
        _REST_HOST       => 'https://sysnews.ncsu.edu',
        _NOCACHE         => undef,
    };

    $params{'USER_NAME'} = ( getpwuid $> );

    if ( defined $params{'FW_TYPE'}
        && $params{'FW_TYPE'} =~ /\A(ipset|iptables|ipfw|route)\z/ )
    {
        $self->{'_FW_TYPE'} = $1;
    }

    # otherwise, try to guess the right type to use
    elsif ( -x "/usr/sbin/ipset" && -x "/sbin/iptables" ) {
        $self->{'_FW_TYPE'} = 'ipset';
    }
    elsif ( -x "/sbin/iptables" ) {
        $self->{'_FW_TYPE'} = "iptables";
    }
    elsif ( -x "/sbin/ipfw" ) {
        $self->{'_FW_TYPE'} = "ipfw";
    }
    elsif ( -x "/usr/sbin/route" ) {
        $self->{'_FW_TYPE'} = "route";
    }
    else {
        $self->{'_FW_TYPE'} = "unknown";
    }

    $self->{'_DB_TYPE'}   = $params{'DB_TYPE'}   if ( $params{'DB_TYPE'} );
    $self->{'_DB_USER'}   = $params{'DB_USER'}   if ( $params{'DB_USER'} );
    $self->{'_DB_PASS'}   = $params{'DB_PASS'}   if ( $params{'DB_PASS'} );
    $self->{'_DB_HOST'}   = $params{'DB_HOST'}   if ( $params{'DB_HOST'} );
    $self->{'_DB_NAME'}   = $params{'DB_NAME'}   if ( $params{'DB_NAME'} );
    $self->{'_REST_HOST'} = $params{'REST_HOST'} if ( $params{'REST_HOST'} );

    $self->{'_USER_NAME'} = $params{'USER_NAME'} if ( $params{'USER_NAME'} );

    $self->{'_VERBOSE'} = $params{'VERBOSE'};
    $self->{'_NOCACHE'} = $params{'NOCACHE'};

    bless( $self, $class );
    return $self;
}

=head2 db_connect

 Usage    : $bt->db_connect;
 Purpose  : Opens a connection to the MySQL database.
 Returns  : A DBI database handle
 Argument : None
 Comments : The username, password, and database names are setup in new()
          : Caches the handle in the object.
 
=cut

sub db_connect {
    my $self = shift;

    # handle caching for fun and profit
    return $self->{_DBHANDLE} if ( defined $self->{_DBHANDLE} );

    my $DBSERVER
        = "DBI:mysql:database=$self->{'_DB_NAME'};host=$self->{'_DB_HOST'}";
    my $dbh
        = DBI->connect( $DBSERVER, $self->{'_DB_USER'}, $self->{'_DB_PASS'} );
    if ( !$dbh ) {
        die("Unable to connect to MySQL database.\n");
    }
    $self->{_DB_TYPE}  = "mysql";
    $self->{_DBHANDLE} = $dbh;
    return $dbh;
}

=head2 get_whitelist

 Usage    : $bt->get_whitelist('category');
 Purpose  : Get a list of the currently whitelisted IP/CIDR addresses matching the supplied pattern.  If no category is supplied, all categories will be matched.
 Returns  : A cidr object containing the whitelisted CIDR/IP addresses
 Argument : 
   category => a string representing the category pattern to match against

=cut

sub get_whitelist {
    my $self     = shift;
    my $category = shift;

    if ($category) {
        $self->{_WLCIDR} = undef;
        $self->{_DBCIDR} = undef
            if ( defined $self->{_CAT} && $self->{_CAT} ne $category );
        $self->{_CAT} = $category;

    }
    else {
        $category = $self->{_CAT} if ( defined $self->{_CAT} );
    }
    return $self->{_WLCIDR} if ( defined $self->{_WLCIDR} );

    # use cache files unless forbidden
    if ( !$self->{_NOCACHE} ) {
        my $cidr = $self->get_static_whites($category);
        return $cidr if ( defined $cidr );
    }

    # WARNING: do not write verbose messages here. Puppet may grab them!
    my $cidr    = Net::CIDR::Lite->new();
    my %wl_hash = ();
    my $list    = $self->list_blocks(
        {   expired   => 0,
            whitelist => 1,
            category  => $category,
        }
    );
    foreach my $r (@$list) {
        $wl_hash{ $r->{ipaddress} } = '1';
    }

    foreach my $k ( sort keys %wl_hash ) {
        if ( $k =~ m/\d+\/\d+/ ) {
            $cidr->add($k);
        }
        else {
            $cidr->add_ip($k);
        }
    }

    $self->{_WLCIDR} = $cidr;
    return $cidr;
}

=head2 get_static_whites

 Usage    : $bt->get_static_whites('category');
 Purpose  : Same function as get_whitelist, except that reads from a pre-cached list file via http rather than taxing the database with repeated queries.
 Returns  : A cidr object containing the whitelisted CIDR/IP addresses
 Argument : 
   category => a string representing the category pattern to match against

=cut

sub get_static_whites {
    my $self     = shift;
    my $category = shift;

    if ($category) {
        $self->{_WLCIDR} = undef;
        $self->{_DBCIDR} = undef
            if ( defined $self->{_CAT} && $self->{_CAT} ne $category );
        $self->{_CAT} = $category;

    }
    else {
        $category = $self->{_CAT} if ( defined $self->{_CAT} );
    }
    return $self->{_WLCIDR} if ( defined $self->{_WLCIDR} );

    my $cidr = Net::CIDR::Lite->new();

    # FIXME: get JSON object @$list from a URL
    # one day the other end of this should be a real REST
    # for now, we just run a cron to leave static JSON files
    # and do a simple LWP grab from this end
    $category = 'undef' if ( !defined $category || $category eq '' );
    $category =~ s/[^\w\-]//g;    # taint fix, just in case
    my $url
        = $self->{'_REST_HOST'}
        . '/rest/badtraffic/static/'
        . $category
        . '.wl.json';
    my $ua  = LWP::UserAgent->new();
    my $req = HTTP::Request->new( GET => $url );
    my $res = $ua->request($req);
    if ( !$res->is_success ) {

        # on failure, fall back to db call
        print STDERR "get_static_whites($category) failed: ",
            $res->message, "\n"
            if ( $self->{'_VERBOSE'} );
        return;
    }
    my $list = JSON->new()->decode( $res->content );

    # WARNING: do not write verbose messages here. Puppet may grab them!
    foreach my $ip (@$list) {
        $cidr->add_any($ip);
    }

    $self->{_WLCIDR} = $cidr;
    return $cidr;
}

=head2 wl_add

 Usage    : $bt->wl_add("address", "reason", "days");
 Purpose  : Add an IP or CIDR address to the whitelist.
 Returns  : Nothing
 Argument : 
   address => IP address or CIDR address as a string
   reason => textual description of the entry
   days => number of days the block will last
 Comment  : The first two arguments (address and reason) are required.  The days argument will be set to 3650 (10 years) unless an argument is given.
 
=cut

sub wl_add {
    my $self = shift;

    my %args = @_;

    my $block    = undef;
    my $reason   = undef;
    my $days     = "1";
    my $category = undef;
    my $hostname = hostname();

    if ( $args{'address'} )  { $block    = $args{'address'} }
    if ( $args{'reason'} )   { $reason   = $args{'reason'} }
    if ( $args{'days'} )     { $days     = $args{'days'} }
    if ( $args{'category'} ) { $category = $args{'category'} }

    unless ($block) {
        warn "WARNING: An address to be whitelisted must be specified\n";
        return 1;
    }
    unless ($reason) {
        warn "WARNING: A reason to whitelist $block must be specified\n";
        return 1;
    }
    unless ( defined $days ) { $days = '1' }

    unless ( $block =~ m/\/\d/ ) {
        $block .= "/32";
    }

    my $dbh = $self->db_connect();

    my $SQL
        = "INSERT INTO blocked_hosts (ipaddress, reason, category, date_blocked, date_expired, reporting_host, whitelisted) VALUES (?, ?, ?, NOW(), NOW()+Interval ? day, ?, '1')";

    print "Adding $block to whitelist\n" if ( $self->{'_VERBOSE'} );
    my $sth = $dbh->prepare($SQL);
    $sth->execute( "$block", "$reason", "$category", "$days", "$hostname" );
    $self->{_WLCIDR}->add($block) if ( defined $self->{_WLCIDR} );

    my $fw_cidr = $self->get_fw_blocks();
    my $wl_cidr = $self->get_whitelist();

    my @fw_cidr_list = $fw_cidr->list;
    my @wl_cidr_list = $wl_cidr->list;

    for my $e (@fw_cidr_list) {
        if ( $wl_cidr->find($e) ) {
            print
                "$e is blocked in the $self->{_FW_TYPE} firewall, but is whitelisted!\n";
            $self->fw_unblock($e);
        }
    }
}

=head2 wl_remove

 Usage    : $bt->wl_remove("address", "category");
 Purpose  : Remove an IP or CIDR address from the whitelist.
 Returns  : Nothing
 Argument : address  => IP address or CIDR address as a string
          : category => only remove from category, if given
 
=cut

sub wl_remove {
    my $self     = shift;
    my $block    = shift;
    my $category = shift;

    unless ( $block =~ m/\/\d/ ) {
        $block .= "/32";
    }
    unless ($block) {
        warn "WARNING: No IP address was specified to unblock\n";
        return 1;
    }
    print "Removing $block from whitelist\n" if ( $self->{'_VERBOSE'} );

    my $SQL = "UPDATE blocked_hosts SET expired = '1' "
        . "WHERE ipaddress = ? AND whitelisted = '1'";
    my @params = ($block);
    if ($category) {
        $SQL .= " AND category=?";
        push @params, $category;
    }

    my $dbh = $self->db_connect();
    my $sth = $dbh->prepare($SQL);
    $sth->execute(@params);
}

=head2 get_fw_blocks

 Usage    : $bt->get_fw_blocks();
 Purpose  : Get list of active blocks on the local system's firewall.
 Returns  : A CIDR object containing the blocked CIDR/IP addresses
 
=cut

sub get_fw_blocks {
    my $self = shift;

    unless ( $self->{_USER_NAME} eq "root" ) {
        warn "Only root can change firewall rules\n";
        return 1;
    }

    my %hash = ();
    my $cidr = Net::CIDR::Lite->new();

    if ( $self->{_FW_TYPE} eq "ipset" ) {
        if ( !-x "/usr/sbin/ipset" ) { return $cidr }
        print "Listing current ipset badtraffic blocks\n"
            if ( $self->{'_VERBOSE'} );
        my @result = `/usr/sbin/ipset list badtraffic`;
        foreach my $line (@result) {
            if ( $line =~ m/^(\d|\.)/ ) {
                $hash{$line} = '1';
            }
        }
    }

    if ( $self->{_FW_TYPE} eq "iptables" ) {
        if ( !-x "/sbin/iptables-save" ) { return $cidr }
        my $iptable  = {};
        my @IPTABLES = `/sbin/iptables-save | grep AUTO`;
        $self->{_IPTABLE_COUNT} = 0;
        foreach my $line (@IPTABLES) {
            my $block;
            if ( $line =~ m{INPUT-AUTO} ) {
                ($block) = ( $line =~ m{-s\s+([\d./]+)} );
            }
            elsif ( $line =~ m{OUTPUT-AUTO} ) {
                ($block) = ( $line =~ m{-d\s+([\d./]+)} );
            }
            next if ( !$block );
            $block .= '/32' if ( $block !~ m{/\d+\z} );
            push @{ $iptable->{$block} }, $line;
            $self->{_IPTABLE_COUNT}++;
            $hash{$block} = '1';
        }
        $self->{_IPTABLE_CACHE} = $iptable;
    }

    if ( $self->{_FW_TYPE} eq "ipfw" ) {
        if ( !-x "/sbin/ipfw" ) { return $cidr }
        my $result   = `sysctl  -w net.inet.ip.fw.autoinc_step=5`;
        my @IPTABLES = `/sbin/ipfw list`;
        foreach my $line (@IPTABLES) {
            if ( $line =~ m/(\d+) deny ip from (.+) to me/ ) {
                my $addy = $2;
                unless ( $addy =~ m/\/\d/ ) { $addy .= "/32" }
                $hash{$addy} = '$1';
            }

            if ( $line =~ m/(\d+) deny ip from me to (.+)/ ) {
                my $addy = $2;
                unless ( $addy =~ m/\/\d/ ) { $addy .= "/32" }
                $hash{$addy} = '$1';
            }
        }
    }

    if ( $self->{_FW_TYPE} eq "route" ) {
        if ( !-x "/usr/bin/netstat" ) { return $cidr }
        my @NETSTAT = `/usr/bin/netstat -rnv`;
        foreach my $line (@NETSTAT) {
            if ( $line =~ m/^127.0.0.1/ ) {next}
            if ( $line =~ m/^((\d+|\.)+) +((\d+|\.)+) +127.0.0.1/ ) {

                #print "converting $1 $3 into CIDR\n";
                my $tmp = $self->netmask2cidr( "$1", "$3" );
                $hash{$tmp} = '1';
            }
        }
    }

    foreach my $k ( sort keys %hash ) {
        if ( $k =~ m/\d+\/\d+/ ) {
            $cidr->add($k);
        }
        else {
            $cidr->add_ip($k);
        }
    }
    return $cidr;
}

=head2 fw_block

 Usage    : $bt->fw_block("address");
 Purpose  : Block an address on the local system's firewall.
 Arguments: The address or network to be blocked.
 Returns  : Nothing
 
=cut

sub fw_block {
    my $self = shift;

    unless ( $self->{_USER_NAME} eq "root" ) {
        warn "Only root can change firewall rules\n";
        return 1;
    }

    my $block  = shift;
    my $result = undef;
    my $cidr   = $self->get_whitelist();

    if ( $cidr->find($block) ) {
        warn
            "WARNING: $block is whitelisted.  Aborting $self->{_FW_TYPE} block.\n";
        return 1;
    }

    unless ($block) {
        warn "WARNING: No IP address was specified to block\n";
        return 1;
    }

    if ( $self->{_FW_TYPE} eq "iptables" ) {
        if ( !-x "/sbin/iptables" ) { return 1 }

        # make sure we stay consistent with db_block,
        # a single IP address must be listed as ip/32.
        if ( $block !~ m{/\d} ) {
            $block .= "/32";
        }

        # check the cache to make sure we don't add duplicate rules
        my $rules = $self->{_IPTABLE_CACHE}->{$block};
        if ( defined $rules && @$rules > 0 ) {
            my $count = @$rules;
            if ( $count > 2 ) {    # should be one pair of rules per range

                # toomany? clear it all out and readd one rule pair
                $self->fw_unblock($block);
            }
            else {

                # quietly skip existing rule pair
                warn "WARNING: block $block refused because "
                    . "$count rules already exist\n"
                    if ( $self->{'_VERBOSE'} );
                return;
            }
        }

        my $cmd;
        print "Blocking $block from iptables INPUT-AUTO chain\n"
            if ( $self->{'_VERBOSE'} );
        $cmd = "-A INPUT-AUTO -s $block -j DROP -m comment "
            . "--comment \"200 badtraffic INPUT-AUTO block rule for $block\"";
        $result = `/sbin/iptables $cmd`;
        print "$cmd resulted in:\n", $result
            if ( $result && $self->{'_VERBOSE'} );
        push @{ $self->{_IPTABLE_CACHE}->{$block} }, $cmd;

        print "Blocking $block from iptables OUTPUT-AUTO chain\n"
            if ( $self->{'_VERBOSE'} );
        $cmd = "-A OUTPUT-AUTO -d $block -j DROP -m comment "
            . "--comment \"200 badtraffic OUTPUT-AUTO block rule for $block\"";
        $result = `/sbin/iptables $cmd`;
        print "$cmd resulted in:\n", $result
            if ( $result && $self->{'_VERBOSE'} );
        push @{ $self->{_IPTABLE_CACHE}->{$block} }, $cmd;
    }

    if ( $self->{_FW_TYPE} eq "ipfw" ) {
        if ( !-x "/sbin/ipfw" ) { return 1 }
        print "Blocking $block from ipfw\n" if ( $self->{'_VERBOSE'} );
        $result = `/sbin/ipfw add deny ip from $block to me`;
        print $result if ( $result && $self->{'_VERBOSE'} );
        $result = `/sbin/ipfw add deny ip from me to $block`;
        print $result if ( $result && $self->{'_VERBOSE'} );
    }

    if ( $self->{_FW_TYPE} eq "route" ) {
        if ( !-x "/usr/sbin/route" ) { return 1 }
        print "Blocking $block from route\n" if ( $self->{'_VERBOSE'} );
        $result = `/usr/sbin/route add $block 127.0.0.1 -blackhole`;
        print $result if ( $result && $self->{'_VERBOSE'} );
    }

    if ( $self->{_FW_TYPE} eq "ipset" ) {
        if ( !-x "/usr/sbin/ipset" ) { return 1 }
        print "Blocking $block from ipset badtraffic \n"
            if ( $self->{'_VERBOSE'} );
        $result = `/usr/sbin/ipset add badtraffic $block -exist`;
    }

}

=head2 fw_unblock

 Usage    : $bt->fw_unblock("address");
 Purpose  : Removed an active block on the local system's firewall.
 Arguments: The address or network to be unblocked.
 Returns  : Nothing
 
=cut

sub fw_unblock {
    my $self = shift;

    unless ( $self->{_USER_NAME} eq "root" ) {
        warn "Only root can change firewall rules\n";
        return 1;
    }
    my $block = shift;
    unless ($block) {
        warn "No IP address was specified to unblock\n";
        return 1;
    }

    if ( $self->{_FW_TYPE} eq "iptables" ) {
        if ( !-x "/sbin/iptables" ) {next}

        # always use CIDR addresses
        $block .= '/32' if ( $block !~ m{/\d+\z} );

        print "######## Unblocking $block from iptables\n"
            if ( $self->{'_VERBOSE'} );

        my $rules = $self->{_IPTABLE_CACHE}->{$block};

        # try cache to clear existing rules
        if ( defined $rules && @$rules ) {
            foreach my $rule (@$rules) {
                $rule =~ s/\A-A/-D/;    # change add to delete
                my $cmd = "/sbin/iptables $rule";
                print "Running '$cmd':\n" if ( $self->{'_VERBOSE'} );
                my $result = `$cmd`;
                print $result if ( $result && $self->{'_VERBOSE'} );
            }

            # cleanup cache
            delete $self->{_IPTABLE_CACHE}->{$block};
        }

        # if we have a cache but no match, its probably the CIDR bug
        elsif ( $self->{_IPTABLE_COUNT} > 0 ) {
            warn "WARNING: possible CIDR compact range bug on $block\n"
                if ( $self->{'_VERBOSE'} );
        }

        # otherwise, make a best guess on the rules to remove
        else {
            my $cmd;
            my $result = undef;
            $cmd
                = "/sbin/iptables -D INPUT-AUTO -s $block -j DROP -m comment "
                . "--comment \"200 badtraffic INPUT-AUTO block rule for $block\"";
            print "Running '$cmd':\n" if ( $self->{'_VERBOSE'} );
            $result = `$cmd`;
            print $result if ( $result && $self->{'_VERBOSE'} );
            $cmd
                = "/sbin/iptables -D OUTPUT-AUTO -d $block -j DROP -m comment "
                . "--comment \"200 badtraffic OUTPUT-AUTO block rule for $block\"";
            print "Running '$cmd':\n" if ( $self->{'_VERBOSE'} );
            $result = `$cmd`;
            print $result if ( $result && $self->{'_VERBOSE'} );
        }
    }
    if ( $self->{_FW_TYPE} eq "ipfw" ) {
        if ( $block =~ m/(.+)\/32/ ) { $block = $1 }
        my @result = `/sbin/ipfw list | grep $block`;
        foreach my $line (@result) {
            if ( $line =~ m/(\d+) deny ip from (.*)/ ) {
                print "Unblocking $2 from ipfw (rule #$1)\n"
                    if ( $self->{'_VERBOSE'} );
                `/sbin/ipfw delete $1`;
            }
        }
    }

    if ( $self->{_FW_TYPE} eq "route" ) {

        my @tmp = $self->cidr2netmask($block);

        print
            "Unblocking $block from route\n /usr/sbin/route delete $tmp[0] 127.0.0.1 $tmp[1]\n"
            if ( $self->{'_VERBOSE'} );
        my @result = `/usr/sbin/route delete $tmp[0] 127.0.0.1 $tmp[1]`;
    }

    if ( $self->{_FW_TYPE} eq "ipset" ) {
        if ( !-x "/usr/sbin/ipset" ) { return 1 }
        print "Unblocking $block from ipset badtraffic \n"
            if ( $self->{'_VERBOSE'} );
        my $result = `/usr/sbin/ipset del badtraffic $block -exist`;
        print $result if ( $result && $self->{'_VERBOSE'} );
    }

}

=head2 fw_insanity

 Usage    : if ($bt->fw_insanity()) { ... }
 Purpose  : Check the loaded firewall rules for inconsistencies.
 Arguments: None.
 Returns  : True if insanity corrections were made.
 
=cut

sub fw_insanity {
    my $self = shift;

    if ( $self->{_FW_TYPE} eq "iptables" ) {
        if ( !$self->{_IPTABLE_COUNT} ) {
            warn "fw_insanity called on empty firewall ruleset\n"
                if ( $self->{'_VERBOSE'} );
            return;
        }
        my $fixes = 0;
        foreach my $range ( keys %{ $self->{_IPTABLE_CACHE} } ) {
            my $count = @{ $self->{_IPTABLE_CACHE}->{$range} };
            if ( $count > 2 ) {
                warn "INSANITY: found $count FW rules for $range\n"
                    if ( $self->{'_VERBOSE'} );
                $fixes++;
                $self->fw_unblock($range);
            }
        }
        return $fixes;
    }

    else {
        warn "No insanity check for $self->{_FW_TYPE}\n"
            if ( $self->{'_VERBOSE'} );
        return;
    }
}

=head2 db_block

 Usage    : $bt->db_block("address", "reason", "days");
 Purpose  : Add a CIDR address to the blacklist.
 Returns  : Nothing
 Argument : 
   address => IP address or CIDR address as a string
   reason => textual description of the entry
   days => number of days the block will last
 Comment  : The first two arguments (address and reason) are required.  The days argument will be set to 1 day unless an argument is given.
 
=cut

sub db_block {
    my $self = shift;
    my %args = @_;

    my $block    = undef;
    my $reason   = undef;
    my $days     = "1";
    my $category = undef;
    my $hostname = hostname();

    if ( $args{'address'} )  { $block    = $args{'address'} }
    if ( $args{'reason'} )   { $reason   = $args{'reason'} }
    if ( $args{'days'} )     { $days     = $args{'days'} }
    if ( $args{'category'} ) { $category = $args{'category'} }

    unless ($block) {
        warn "WARNING: An address to be blocked must be specified\n";
        return 1;
    }
    unless ($reason) {
        warn "WARNING: A reason to block $block must be specified\n";
        return 1;
    }
    unless ( defined $days ) { $days = '1' }

    unless ( $block =~ m/\/\d/ ) {
        $block .= "/32";
    }

    print "Adding a database block for $block\n" if ( $self->{'_VERBOSE'} );

    my $SQL
        = "INSERT INTO blocked_hosts (ipaddress, reason, date_blocked, date_expired, reporting_host, category) VALUES (?, ?, NOW(), NOW()+Interval ? day, ?, ?)";
    my $dbh = $self->db_connect();
    my $sth = $dbh->prepare($SQL);
    $sth->execute( "$block", "$reason", "$days", "$hostname", "$category" );
    $self->{_DBCIDR}->add($block) if ( defined $self->{_DBCIDR} );
}

=head2 db_unblock

 Usage    : $bt->db_unblock("address", "category");
 Purpose  : Remove an IP or CIDR address from the blacklist.
 Returns  : Nothing
 Argument : address => IP address or CIDR address as a string
          : category => remove from given category, if provided
 
=cut

sub db_unblock {
    my $self     = shift;
    my $block    = shift;
    my $category = shift;

    unless ( $block =~ m/\/\d/ ) {
        $block .= "/32";
    }
    unless ($block) {
        warn "WARNING: No IP address was specified to unblock\n";
        return 1;
    }

    print "Removing database blocks for $block\n" if ( $self->{'_VERBOSE'} );

    my $SQL = "UPDATE blocked_hosts SET expired = '1' "
        . "WHERE ipaddress = ? AND whitelisted = '0'";
    my @params = ($block);
    if ($category) {
        $SQL .= " AND category=?";
        push @params, $category;
    }

    my $dbh = $self->db_connect();
    my $sth = $dbh->prepare($SQL);
    $sth->execute(@params);
}

=head2 db_block_expire

 Usage    : $bt->db_block_expire();
 Purpose  : Sets expired flag for all entries that have expired.
 Returns  : Nothing
 Comments : This function is always called when get_db_blocks() is called.
 
=cut

sub db_block_expire {
    my $self = shift;

    # processes using cache don't need to run this
    return if ( !$self->{_NOCACHE} );

    # this really only needs to run once/session
    return if ( $self->{_RANEXPIRE} );

    my $dbh = $self->db_connect();
    print "Expiring old database blocks\n" if ( $self->{'_VERBOSE'} );
    my $SQL
        = "UPDATE blocked_hosts SET expired = '1' WHERE date_expired < NOW() AND expired != '1'";
    my $sth = $dbh->prepare($SQL);
    $sth->execute();
    $self->{_RANEXPIRE} = 1;
}

=head2 get_cat_list 

 Usage    : my @cats = $bt->get_cat_list('category');
 Purpose  : Get list of categories that are included in the rules for the given category.
 Returns  : An array of included categories, if any are found.
 Argument : a string for the parent category

=cut

sub get_cat_list {
    my $self     = shift;
    my $category = shift;

    return () if ( !$category );
    my @res = ();

    my $SQL = "select include from cat_includes where category = ?";
    my $dbh = $self->db_connect();
    my $sth = $dbh->prepare($SQL);
    $sth->execute($category);
    while ( my $r = $sth->fetchrow_hashref() ) {
        push @res, $r->{include};
    }
    return @res;
}

=head2 cat_add

 Usage    : $res = $bt->cat_add($parent, $child);
 Purpose  : Add a child category list to a parent category.
 Returns  : 1 if the child was added, undef on failure.
 Argument : $parent = primary category string
          : $child  = included category string

=cut 

sub cat_add {
    my $self   = shift;
    my $parent = shift || return;
    my $child  = shift || return;

    # avoid duplicate entries
    my %known = map { $_ => 1 } $self->get_cat_list($parent);
    return 1 if ( $known{$child} );

    my $SQL = "insert into cat_includes (category, include) VALUES (?, ?)";
    my $dbh = $self->db_connect();
    my $sth = $dbh->prepare($SQL);
    $sth->execute( $parent, $child ) or return;
    return 1;
}

=head2 cat_remove

 Usage    : $res = $bt->cat_remove($parent, $child);
 Purpose  : Remove a child category list from a parent category.
 Returns  : 1 if the child was removed, undef on failure.
 Argument : $parent = primary category string
          : $child  = included category string

=cut 

sub cat_remove {
    my $self   = shift;
    my $parent = shift || return;
    my $child  = shift || return;

    # anything to remove?
    my %known = map { $_ => 1 } $self->get_cat_list($parent);
    return 1 if ( !$known{$child} );

    my $SQL = "delete from cat_includes where category=? and include=?";
    my $dbh = $self->db_connect();
    my $sth = $dbh->prepare($SQL);
    $sth->execute( $parent, $child ) or return;
    return 1;
}

=head2 get_db_blocks

 Usage    : $bt->get_db_blocks('category');
 Purpose  : Get list of active blacklist entries from the database matching the supplied pattern.  If no category is supplied, all categories will be matched.
 Returns  : A cidr object containing the blacklisted CIDR/IP addresses
 Argument : 
   category => a string representing the category pattern to match against

=cut

sub get_db_blocks {
    my $self     = shift;
    my $category = shift;

    if ($category) {
        $self->{_DBCIDR} = undef;
        $self->{_WLCIDR} = undef
            if ( defined $self->{_CAT} && $self->{_CAT} ne $category );
        $self->{_CAT} = $category;
    }
    else {
        $category = $self->{_CAT} if ( defined $self->{_CAT} );
    }
    return $self->{_DBCIDR} if ( defined $self->{_DBCIDR} );

    # use cache files unless forbidden
    if ( !$self->{_NOCACHE} ) {
        my $cidr = $self->get_static_blocks($category);
        return $cidr if ( defined $cidr );
    }

    my $cidr = Net::CIDR::Lite->new();
    my %hash = ();
    my $list = $self->list_blocks(
        {   expired   => 0,
            blacklist => 1,
            category  => $category,
        }
    );

    # WARNING: puppet expects to see these messages
    foreach my $r (@$list) {
        print "$r->{ipaddress} is currently blocked in the database\n"
            if ( $self->{'_VERBOSE'} );
        $hash{ $r->{ipaddress} } = '1';
    }

    foreach my $k ( sort keys %hash ) {
        if ( $k =~ m/\d+\/\d+/ ) {
            $cidr->add($k);
        }
        else {
            $cidr->add_ip($k);
        }
    }

    $self->{_DBCIDR} = $cidr;
    return $cidr;
}

=head2 get_static_blocks

 Usage    : $bt->get_static_blocks('category');
 Purpose  : Same function as get_db_blocks, except that reads from a pre-cached list file via http rather than taxing the database with repeated queries.
 Returns  : A cidr object containing the blacklisted CIDR/IP addresses
 Argument : 
   category => a string representing the category pattern to match against

=cut

sub get_static_blocks {
    my $self     = shift;
    my $category = shift;

    if ($category) {
        $self->{_DBCIDR} = undef;
        $self->{_WLCIDR} = undef
            if ( defined $self->{_CAT} && $self->{_CAT} ne $category );
        $self->{_CAT} = $category;
    }
    else {
        $category = $self->{_CAT} if ( defined $self->{_CAT} );
    }
    return $self->{_DBCIDR} if ( defined $self->{_DBCIDR} );

    my $cidr = Net::CIDR::Lite->new();

    # FIXME: get JSON object @$list from a URL
    # one day the other end of this should be a real REST
    # for now, we just run a cron to leave static JSON files
    # and do a simple LWP grab from this end
    $category = 'undef' if ( !defined $category || $category eq '' );
    $category =~ s/[^\w\-]//g;    # taint fix, just in case
    my $url
        = $self->{'_REST_HOST'}
        . '/rest/badtraffic/static/'
        . $category
        . '.bl.json';
    my $ua  = LWP::UserAgent->new();
    my $req = HTTP::Request->new( GET => $url );
    my $res = $ua->request($req);
    if ( !$res->is_success ) {

        # on failure, fall back to db call
        print STDERR "get_static_block($category) failed: ",
            $res->message, "\n"
            if ( $self->{'_VERBOSE'} );
        return;
    }
    my $list = JSON->new()->decode( $res->content );

    # WARNING: puppet expects to see these messages
    foreach my $ip (@$list) {
        print "$ip is currently blocked in the database\n"
            if ( $self->{'_VERBOSE'} );
        $cidr->add_any($ip);
    }

    $self->{_DBCIDR} = $cidr;
    return $cidr;
}

=head2 get_db_block_reasons

 Usage    : my %dbreasons = $bt->get_db_block_reasons('category');
 Purpose  : Get list of active blacklist entries from the database matching the supplied pattern.  If no category is supplied, all categories will be matched.
 Returns  : A hash keyed by IP/CIDR, with reason strings as values
 Argument : 
   category => a string representing the category pattern to match against

=cut

sub get_db_block_reasons {
    my $self     = shift;
    my $category = shift;

    $self->db_block_expire();
    my %hash = ();
    my $list = $self->list_blocks(
        {   expired   => 0,
            blacklist => 1,
            category  => $category,
        }
    );
    foreach my $r (@$list) {
        print "$r->{ipaddress} has db block for '$r->{reason}'\n"
            if ( $self->{'_VERBOSE'} );
        $hash{ $r->{ipaddress} } = $r->{reason};
    }

    return %hash;
}

=head2 get_recent_blocks

 Usage    : $bt->get_recent_blocks();
 Purpose  : Get list of all IP Addresses blacklisted in the last 30 days.
 Returns  : A hash containg the blacklisted CIDR/IP addresses, with the IP being the key, and the number of entries being the value.
 Argument : The number of days to look back through.  Defaults to 30.

=cut

sub get_recent_blocks {
    my $self = shift;
    my $days = shift || 30;

    my $cidr = Net::CIDR::Lite->new();
    my %hash = ();

    # check file cache first
    if ( !$self->{_NOCACHE} ) {
        my $obj = $self->get_recent_static($days);
        if ( defined $obj->{list} ) {
            foreach my $ip ( @{ $obj->{list} } ) {
                $cidr->add_any($ip);
            }
            return $cidr;
        }
    }

    my $list = $self->list_blocks(
        {   expired => 0,
            recent  => $days,
        }
    );
    foreach my $r (@$list) {
        print "$r->{ipaddress} has been bad in the last $days days\n"
            if ( $self->{'_VERBOSE'} );
        $hash{ $r->{ipaddress} }++;
    }
    foreach my $k ( sort keys %hash ) {
        if ( $k =~ m/\d+\/\d+/ ) {
            $cidr->add($k);
        }
        else {
            $cidr->add_ip($k);
        }
    }
    return $cidr;
}

=head2 get_recent_blocks_count

 Usage    : $bt->get_recent_blocks();
 Purpose  : Get list of all IP Addresses blacklisted in the last 30 days.
 Returns  : A hash containg the blacklisted CIDR/IP addresses, with the IP being the key, and the number of entries being the value.
 Argument : The number of days to look back through.  Defaults to 30.

=cut

sub get_recent_blocks_count {
    my $self = shift;
    my $days = shift;
    unless ($days) { $days = "30"; }

    # check file cache first
    if ( !$self->{_NOCACHE} ) {
        my $obj = $self->get_recent_static($days);
        return %{ $obj->{count} } if ( defined $obj->{count} );
    }

    my %hash = ();
    my $dbh  = $self->db_connect();
    my $SQL
        = "SELECT ipaddress FROM blocked_hosts WHERE date_blocked > NOW()-Interval $days day AND expired = '1'";
    my $sth = $dbh->prepare($SQL);
    $sth->execute();

    while ( my $r = $sth->fetchrow_hashref() ) {
        print "$r->{ipaddress} has been bad in the last $days days\n"
            if ( $self->{'_VERBOSE'} );
        $hash{ $r->{ipaddress} }++;
    }
    return %hash;
}

=head2 get_recent_static

 Usage    : my $obj = $bt->get_recent_blocks();
 Purpose  : Retrive list of blocked IPs and counts by IP from cached files
 Returns  : An object containing { 'list' => [ array of ips ], 'count' => 'counts by ip' }
 Argument : The number of days to look back through.  Defaults to 30.

=cut

sub get_recent_static {
    my $self = shift;
    my $days = shift;
    $days =~ s/\D+//g;
    $days = 30 if ( !$days );
    my $cachevar = '_DC_' . $days;

    # check for previous call
    return $self->{$cachevar} if ( defined $self->{$cachevar} );

    my $url
        = $self->{'_REST_HOST'}
        . '/rest/badtraffic/static/recent.'
        . $days . '.json';
    my $ua  = LWP::UserAgent->new();
    my $req = HTTP::Request->new( GET => $url );
    my $res = $ua->request($req);
    if ( !$res->is_success ) {

        # on failure, fall back to db calls
        print STDERR "get_recent_static($days) failed: ", $res->message, "\n"
            if ( $self->{'_VERBOSE'} );
        return;
    }
    print "Read ", length( $res->content ), " bytes from recent $days cache\n"
        if ( $self->{'_VERBOSE'} );
    my $obj = JSON->new()->decode( $res->content );
    $self->{$cachevar} = $obj;
    return $obj;

}

=head2 check_ipset_setup

 Usage    : $bt->check_ipset_setup();
 Purpose  : Sets up ipset and iptables for use with BadTraffic
 Returns  : Nothing
 
=cut

sub check_ipset_setup {
    my $self = shift;
    if ( $self->{_FW_TYPE} ne "ipset" ) {

        #Setting up ipset firewalls won't help if you aren't running ipset
        return 0;
    }

    unless ( $self->{_USER_NAME} eq "root" ) {
        warn "Only root can change firewall rules\n";
        return 1;
    }

    my @result = ();

    my $action = 1;

    @result = `/usr/sbin/ipset create badtraffic hash:net -exist`;

    @result = `/sbin/iptables -L INPUT -n -v`;
    foreach my $line (@result) {
        if ( $line =~ m/match-set badtraffic src/ ) {

            print "match-set badtraffic jump rule is in place\n"
                if ( $self->{'_VERBOSE'} );
            $action = 0;
        }
    }

    if ($action) {
        print "Adding ipset jump from INPUT\n";
        @result
            = `/sbin/iptables -I INPUT 1 -m set --set badtraffic src -j DROP`;
    }

    @result = `/sbin/iptables -L OUTPUT -n -v`;
    foreach my $line (@result) {
        if ( $line =~ m/match-set badtraffic src/ ) {

            print "match-set badtraffic jump rule is in place\n"
                if ( $self->{'_VERBOSE'} );
            $action = 0;
        }
    }

    if ($action) {
        print "Adding ipset jump from OUTPUT\n";
        @result
            = `/sbin/iptables -I OUTPUT 1 -m set --set badtraffic src -j DROP`;
    }
}

=head2 check_iptables_setup

 Usage    : $bt->check_iptables_setup();
 Purpose  : Sets up iptables for use with BadTraffic
 Returns  : Nothing
 
=cut

sub check_iptables_setup {
    my $self = shift;
    if ( $self->{_FW_TYPE} eq "ipset" ) {
        &check_ipset_setup($self);
    }
    if ( $self->{_FW_TYPE} ne "iptables" ) {

      #Setting up iptables firewalls won't help if you aren't running iptables
        return 0;
    }

    unless ( $self->{_USER_NAME} eq "root" ) {
        warn "Only root can change firewall rules\n";
        return 1;
    }

    my @result = ();

    my $action = 1;

    @result = `/sbin/iptables -L INPUT-AUTO -n -v`;
    foreach my $line (@result) {
        if ( $line =~ m/Chain INPUT-AUTO \((\d+) references\)/ ) {

            #print "INPUT-AUTO jump rule is in place with $1 references\n";
            $action = 0;
        }
    }

    if ($action) {
        print "Adding iptables chain INPUT-AUTO\n";
        @result = `/sbin/iptables -N INPUT-AUTO`;
    }

    $action = 1;
    @result = `/sbin/iptables -L OUTPUT-AUTO -n -v`;
    foreach my $line (@result) {
        if ( $line =~ m/Chain OUTPUT-AUTO \((\d+) references\)/ ) {

            #print "OUTPUT-AUTO jump rule is in place with $1 references\n";
            $action = 0;
        }
    }
    if ($action) {
        print "Adding iptables chain OUTPUT-AUTO\n";
        @result = `/sbin/iptables -N OUTPUT-AUTO`;
    }

    $action = 1;

    @result = `/sbin/iptables -L INPUT -n -v`;
    foreach my $line (@result) {
        if ( $line =~ m/INPUT-AUTO  all +-- +\* +\* +0.0.0.0/ ) {

            #print "INPUT-AUTO jump rule is in place in the INPUT chain\n";
            $action = 0;
        }
    }
    if ($action) {
        print
            "Adding iptables jump rule for INPUT chain jump to INPUT-AUTO\n";
        @result = `/sbin/iptables -I INPUT -j INPUT-AUTO`;
    }

    $action = 1;

    @result = `/sbin/iptables -L OUTPUT -n -v`;
    foreach my $line (@result) {
        if ( $line =~ m/OUTPUT-AUTO  all +-- +\* +\* +0.0.0.0/ ) {

            #print "OUTPUT-AUTO jump rule is in place in the OUTPUT chain\n";
            $action = 0;
        }
    }
    if ($action) {
        print
            "Adding iptables jump rule for OUTPUT chain jump to OUTPUT-AUTO\n";
        @result = `/sbin/iptables -I OUTPUT -j OUTPUT-AUTO`;
    }
}

# internal tables for netmask converts
my %masktobits = (
    "255.255.255.255" => "32",
    "255.255.255.254" => "31",
    "255.255.255.252" => "30",
    "255.255.255.248" => "29",
    "255.255.255.240" => "28",
    "255.255.255.224" => "27",
    "255.255.255.192" => "26",
    "255.255.255.128" => "25",
    "255.255.255.0"   => "24",
    "255.255.254.0"   => "23",
    "255.255.252.0"   => "22",
    "255.255.248.0"   => "21",
    "255.255.240.0"   => "20",
    "255.255.224.0"   => "19",
    "255.255.192.0"   => "18",
    "255.255.128.0"   => "17",
    "255.255.0.0"     => "16",
    "255.254.0.0"     => "15",
    "255.252.0.0"     => "14",
    "255.248.0.0"     => "13",
    "255.240.0.0"     => "12",
    "255.224.0.0"     => "11",
    "255.192.0.0"     => "10",
    "255.128.0.0"     => "9",
    "255.0.0.0"       => "8",
    "254.0.0.0"       => "7",
    "252.0.0.0"       => "6",
    "248.0.0.0"       => "5",
    "240.0.0.0"       => "4",
    "224.0.0.0"       => "3",
    "192.0.0.0"       => "2",
    "128.0.0.0"       => "1",
    "0.0.0.0"         => "0"
);
my %bitstomask = reverse %masktobits;

=head2 netmask2cidr

 Usage    : $bt->netmask2cidr("address", "netmask");
 Purpose  : Convert an IP address and a netmask into a CIDR address
 Arguments: Two strings, an address and a netmask
 Returns  : A string containing the CIDR address
 
=cut

sub netmask2cidr {
    my $self      = shift;
    my $ipaddress = shift;
    my $netmask   = shift;
    my $cidr      = undef;
    $cidr = $ipaddress . "/" . $masktobits{$netmask};
    return $cidr;
}

=head2 cidr2netmask

 Usage    : $bt->cidr2netmask("cidr");
 Purpose  : Convert a CIDR address into an IP address and a netmask
 Arguments: A string containing the CIDR address
 Returns  : An array containing the address and the netmask
 
=cut

sub cidr2netmask {
    my $self = shift;
    my $cidr = shift;
    my @tmp  = split( '/', $cidr );
    $tmp[1] = $bitstomask{ $tmp[1] };
    return @tmp;
}

=head2 delete_rule 

 Usage    : my $res = $bt->delete_rule(id => $id);
 Purpose  : Explicitly remove a whitelist or block rule by db id number.
 Argument : id => database row id number
 Returns  : 1 on success, undef on failure
 Comment  : Expected to be used to remove duplicate rows.

=cut

sub delete_rule {
    my $self   = shift;
    my %params = (@_);

    my $id = $params{id} || return;

    my $dbh = $self->db_connect();
    my $sth = $dbh->prepare('delete from blocked_hosts where id=?');
    return $sth->execute($id);
}

=head2 list_blocks 

 Usage    : my $aref = $bt->list_blocks({param => val, ...});
 Purpose  : General search routine for reading blocks lists from db
 Arguments: A hash reference of search constraints:
          :     whitelist => 1, 
          :     blacklist => 1,
          :     expired   => 0 | 1,
          :     onlycat   => 'catstring',  # matches only that category
          :     category  => 'catstring',  # matches category regex, any/always, and includes
          :     recent    => $days,        # matches rules created in last $days days
 Returns  : An reference to an array of matching block rows.
 Comment  : This is called internally, but can be used for external 
          : database queries if needed.

=cut

sub list_blocks {
    my $self   = shift;
    my $params = shift;
    my $list   = [];

    # build the where statements from the passed params
    my @wheres    = ();
    my @sqlpars   = ();
    my @globalcat = ();
    if ( $params->{whitelist} ) {
        push @wheres,    "whitelisted = 1";
        push @globalcat, 'always';
    }
    elsif ( $params->{blacklist} ) {
        push @wheres,    "whitelisted = 0";
        push @globalcat, 'all';
    }
    else {
        push @globalcat, 'always', 'all';
    }
    if ( defined $params->{expired} && $params->{expired} == 0 ) {
        push @wheres, "date_expired > NOW()";
        push @wheres, "expired != 1";
    }
    elsif ( defined $params->{expired} && $params->{expired} > 0 ) {
        push @wheres, "(date_expired < NOW() OR expired = 1)";
    }
    if ( defined $params->{recent} ) {
        my ($days) = ( $params->{recent} =~ m{(\d+)} );
        push @wheres, "date_blocked > NOW()-Interval $days day";
    }
    if ( defined $params->{iplike} ) {
        push @wheres,  "ipaddress like ?";
        push @sqlpars, $params->{iplike};
    }

    # category statements can get messy
    if ( defined $params->{onlycat} ) {
        push @wheres,  "category = ?";
        push @sqlpars, $params->{onlycat};
    }
    elsif ( defined $params->{category}
        && $params->{category} !~ /\A(all|always)\z/ )
    {
        my $cat     = $params->{category};
        my $likecat = '%' . $cat . '%';
        my $catsql  = "( category like ? ";
        push @sqlpars, $likecat;
        foreach my $glob (@globalcat) {
            $catsql .= "OR category = ? ";
            push @sqlpars, $glob;
        }
        $catsql .= "OR category IN "
            . "(SELECT include FROM cat_includes WHERE category=?))";
        push @sqlpars, $cat;
        push @wheres,  $catsql;
    }
    else {

        # no category restrictions used for all/always/no-cat
    }

    # avoid throwing errors when no where statements were generated
    if ( !@wheres ) {
        warn "cannot search without at least one param\n";
        return;
    }

    $self->db_block_expire();
    my $SQL = "SELECT * FROM blocked_hosts WHERE " . join( ' AND ', @wheres );
    if ( $self->{_VERBOSE} ) {
        print STDERR "list_blocks SQL:\n$SQL\n";
        print STDERR "Params: ", join( ', ', @sqlpars ), "\n";
    }
    my $dbh = $self->db_connect();
    my $sth = $dbh->prepare($SQL);
    $sth->execute(@sqlpars);
    while ( my $r = $sth->fetchrow_hashref() ) {
        push @$list, $r;
    }

    return $list;
}

=head2 range_find

 Usage    : if ( $bt->range_find( $cidr, $address )) { # do }
 Purpose  : Search the given cidr for an address
 Arguments: $cidr = a Net::CIDR::Lite object
          : $address = an IP address or CIDR range string 
 Returns  : 1 if the $address is covered by the $cidr block
 Comments : This is an improved version of $cidr->find() that 
          : works correctly when $address is a range like '1.2.3.0/24'.
          : On the first call, prep_find() is called and the 
          : $cidr ranges are cached. Subsequent calls use the cache. 
          : Call prep_find() explicitly if you need to rebuild the cache.
=cut 

sub range_find {
    my $self = shift;
    my $cidr = shift;
    my $ip   = shift;

    $self->prep_find($cidr) if ( !defined $self->{FINDER} );
    my $bitmask = $self->bitmask($ip);
    my $low     = 0;
    my $high    = $#{ $self->{FINDER} };
    return if ( $high < 0 );
    if ( $high == 0 ) {
        return 1 if ( $bitmask =~ m/\A$self->{FINDER}->[$high]/ );
        return;
    }
    while ( $low < $high ) {
        my $mid = int( ( $low + $high ) / 2 );
        if ( $mid == $low ) {    # down to the last two options
            return 1 if ( $bitmask =~ m/\A$self->{FINDER}->[$low]/ );
            return 1 if ( $bitmask =~ m/\A$self->{FINDER}->[$high]/ );
            return;
        }
        return 1 if ( $bitmask =~ m/\A$self->{FINDER}->[$mid]/ );
        if ( $bitmask lt $self->{FINDER}->[$mid] ) {
            $high = $mid;
        }
        else {
            $low = $mid;
        }
    }
    return;    # should not happen
}

=head2 prep_find

 Usage    : $bt->prep_find( $cidr );
 Purpose  : Build the range_find cache for the given CIDR object.
 Arguments: $cidr = a Net::CIDR::Lite object
 Returns  : nothing

=cut

sub prep_find {
    my $self = shift;
    my $cidr = shift;

    $self->{FINDER} = [];
    foreach my $ip ( $cidr->list() ) {
        push @{ $self->{FINDER} }, $self->bitmask($ip);
    }
}

=head2 bitmask

 Usage    : my $str = $bt->bitmark( $address );
 Purpose  : Build a binary mask string from the given address
 Arguments: $address = an IP address or CIDR range string 
 Returns  : A binary string of the significant bits in the range.
 Comments : These bitmask strings can be compared as regexes
          : to determine if one range belongs to another range.

=cut

sub bitmask {
    my $self = shift;
    my $ip   = shift;
    my $bits = 32;

    if ( $ip =~ m{([\d.]+)/(\d+)} ) {
        ( $ip, $bits ) = ( $1, $2 );
    }
    my $bitstr
        = join( '', unpack( "xB*", Net::CIDR::Lite::_pack_ipv4($ip) ) );
    return substr( $bitstr, 0, $bits );
}

sub network_address {
    my $self = shift;
    my $cidr = shift;

    my @ip_array = $self->cidr2netmask($cidr);

    my @addrarr = split( /\./, $ip_array[0] );
    my ($ipaddress) = unpack( "N", pack( "C4", @addrarr ) );

    my @maskarr = split( /\./, $ip_array[1] );
    my ($netmask) = unpack( "N", pack( "C4", @maskarr ) );

    my $netadd     = ( $ipaddress & $netmask );
    my @netarray   = unpack( "C4", pack( "N", $netadd ) );
    my $netaddress = join( ".", @netarray );

    $cidr = $self->netmask2cidr( $netaddress, $ip_array[1] );

    return "$cidr";

}

sub broadcast_address {
    my $self = shift;
    my $cidr = shift;

    my @ip_array = $self->cidr2netmask($cidr);

    my @addrarr = split( /\./, $ip_array[0] );
    my ($ipaddress) = unpack( "N", pack( "C4", @addrarr ) );

    my @maskarr = split( /\./, $ip_array[1] );
    my ($netmask) = unpack( "N", pack( "C4", @maskarr ) );

    my $netadd     = ( $ipaddress & $netmask );
    my @netarray   = unpack( "C4", pack( "N", $netadd ) );
    my $netaddress = join( ".", @netarray );

    my $bcast = ( $ipaddress & $netmask ) + ( ~$netmask );
    my @bcastarr = unpack( "C4", pack( "N", $bcast ) );
    my $broadcast = join( ".", @bcastarr );

    $cidr = $self->netmask2cidr( $broadcast, $ip_array[1] );

    return "$cidr";

}

sub both_halves {
    my $self    = shift;
    my $address = shift;
    my $bottom  = $self->network_address($address);
    my $top     = $self->broadcast_address($address);

    my @ip = split( "/", $address );
    my $network = $ip[1];
    $network++;

    if ( $network >= 32 ) { $network = '32' }

    #print "looking for $address - splitting into $bottom and $top\n";

    my @top_array = split( "/", $top );
    my $newtop = $top_array[0] . "/" . $network;

    my @bottom_array = split( "/", $bottom );
    my $newbottom = $bottom_array[0] . "/" . $network;

    my @return = ();

    $return[0] = $self->network_address($newtop);
    $return[1] = $self->network_address($newbottom);

    return @return;
}

sub cidr_contains {
    my $self  = shift;
    my $cidr1 = shift;
    my $cidr2 = shift;

    my @host1_array = split( "/", $cidr1 );
    my @host2_array = split( "/", $cidr2 );
    my $host1_check = $host2_array[0] . "/" . $host1_array[1];
    my $network1    = $self->network_address($host1_check);
    if ( $network1 eq $cidr1 ) { return 1 }
    return 0;
}

sub decompose {
    my $self    = shift;
    my $target  = shift;
    my $address = shift;
    my $limit   = shift;
    my %remove  = ();

    #our %decomposed_list = ();
    unless ($limit) { die "no limit given"; }

    #print "  decomposing $address, looking for $target\n";

    my @both = $self->both_halves($address);

    my @newmask = split( '/', $both[0] );

    unless ( $newmask[1] >= "32" ) {

        #print "$both[0] and $target\n";

        if ( $self->cidr_contains( $both[0], $target ) ) {
            $self->decompose( $target, $both[0], $limit );
        }
        else {

            #print "   putting back $both[0]\n";
            $self->{'_DECOMPOSED_LIST'}{ $both[0] } = '1';
        }

        if ( $self->cidr_contains( $both[1], $target ) ) {
            $self->decompose( $target, $both[1], $limit );
        }
        else {

            #print "   putting back $both[1]\n";
            $self->{'_DECOMPOSED_LIST'}{ $both[1] } = '1';
        }

    }
    return %remove;

    #return $self->{'_DECOMPOSED_LIST'};
}

sub cidr_subtract {

    #Does $cidr1 contain $cidr2?
    my $self  = shift;
    my $cidr1 = shift;
    my $cidr2 = shift;

    my @cidr1_list = $cidr1->list;
    my @cidr2_list = $cidr2->list;

    my $saftey_count = @cidr2_list;
    if ( $saftey_count > 200 ) {
        warn
            "Giving Up: cidr_subtract cannot process lists over 200 entries long\n";
        return 1;
    }

    my %cidr1_hash = map { $_ => '1' } @cidr1_list;
    my %cidr2_hash = map { $_ => '1' } @cidr2_list;

    foreach my $host1 (@cidr2_list)
    {    #preserves IP sort order from Net::CIDR::Lite
            #foreach my $host1 (sort keys %cidr2_hash){
            #print "processing cidr2_hash $host1\n";
        $self->{'_DECOMPOSED_LIST'} = undef;
        $self->{'_DECOMPOSED_LIST'} = {};

        foreach my $host2 (
            sort {
                my @a = split /\./, $a;
                my @b = split /\./, $b;
                $a[0] <=> $b[0]
                    or $a[1] <=> $b[1]
                    or $a[2] <=> $b[2];    #or
                                           #$a[3] <=> $b[3];
            } keys %cidr1_hash
            )
        {

            #print " processing cidr1_hash $host2\n";
            my @host1_array = split( "/", $host1 );
            my @host2_array = split( "/", $host2 );

            my $host2_check = $host1_array[0] . "/" . $host2_array[1];
            my $network2    = $self->network_address($host2_check);

            if ( $network2 eq $host2 ) {

     #print "  removing $host2 from cidr1_hash as it conflicts with $host1\n";
                delete $cidr1_hash{$host2};
                my %final_list = ();
                my %remove
                    = $self->decompose( $host1, $host2, $host1_array[1] );

                foreach my $e ( sort keys %{ $self->{'_DECOMPOSED_LIST'} } ) {

                    #print "  adding $e to cidr_hash1\n";
                    $cidr1_hash{$e} = '1';
                }
            }
        }
    }

    foreach my $host1 ( sort keys %cidr2_hash ) {
        foreach my $host2 ( sort keys %cidr1_hash ) {
            my @host1_array = split( "/", $host1 );
            my @host2_array = split( "/", $host2 );

            #print ".";
            my $host1_check = $host2_array[0] . "/" . $host1_array[1];
            my $network1    = $self->network_address($host1_check);
            if ( $network1 eq $host1 ) { delete $cidr1_hash{$host2} }
        }
    }

    my $new_cidr = Net::CIDR::Lite->new();
    foreach my $key ( sort keys %cidr1_hash ) { $new_cidr->add("$key") }
    return $new_cidr;
}

1;
__END__

